/**
 * DevExtreme (ui/selection.js)
 * Version: 16.2.1-16314
 * Build date: Wed Nov 09 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 */
"use strict";
var $ = require("jquery"),
    Class = require("../core/class"),
    gridCore = require("../ui/data_grid/ui.data_grid.core"),
    gridCoreUtils = require("../ui/grid_core/ui.grid_core.utils"),
    baseEqualKeys = gridCoreUtils.equalKeys,
    commonUtils = require("../core/utils/common"),
    dataQuery = require("../data/query"),
    getKeyHash = gridCore.getKeyHash,
    errors = require("../ui/widget/ui.errors");
var getEqualKeys = function(equalByReference) {
    return function(key1, key2) {
        if (equalByReference) {
            if (commonUtils.isObject(key1) && commonUtils.isObject(key2)) {
                return key1 === key2
            }
        }
        return baseEqualKeys(key1, key2)
    }
};
var equalKeys = getEqualKeys();

function SelectionFilterCreator(keyExpr, selectedItemKeys, isSelectAll) {
    this.getFilter = function() {
        return this._filter || this.getExpr() || this.getFunction()
    };
    this.getFunction = function() {
        if (!isFunction()) {
            return
        }
        return functionFilter
    };
    this.getExpr = function() {
        if (isFunction()) {
            return
        }
        var filterExpr;
        for (var i = 0, length = selectedItemKeys.length; i < length; i++) {
            filterExpr = filterExpr || [];
            var filterExprPart, itemKeyValue = selectedItemKeys[i];
            if (i > 0) {
                filterExpr.push(isSelectAll ? "and" : "or")
            }
            if ($.isArray(keyExpr)) {
                filterExprPart = getFilterForCompositeKey(itemKeyValue)
            } else {
                filterExprPart = getFilterForPlainKey(itemKeyValue)
            }
            filterExpr.push(filterExprPart)
        }
        this._filter = filterExpr;
        return filterExpr
    };
    this.getCombinedFilter = function(dataSourceFilter) {
        var filterExpr = this.getExpr(),
            combinedFilter = filterExpr;
        if (isSelectAll && dataSourceFilter) {
            if (filterExpr) {
                combinedFilter = [];
                combinedFilter.push(filterExpr);
                combinedFilter.push(dataSourceFilter)
            } else {
                combinedFilter = dataSourceFilter
            }
        }
        return combinedFilter
    };
    var isFunction = function() {
        return !keyExpr
    };
    var functionFilter = function(item) {
        for (var i = 0; i < selectedItemKeys.length; i++) {
            if (equalKeys(selectedItemKeys[i], item)) {
                return !isSelectAll
            }
        }
        return isSelectAll
    };
    var getFilterForPlainKey = function(keyValue, key) {
        return [key || keyExpr, isSelectAll ? "<>" : "=", keyValue]
    };
    var getFilterForCompositeKey = function(itemKeyValue) {
        var filterExpr = [];
        for (var i = 0, length = keyExpr.length; i < length; i++) {
            if (i > 0) {
                filterExpr.push(isSelectAll ? "or" : "and")
            }
            var currentKeyExpr = keyExpr[i],
                filterExprPart = getFilterForPlainKey(itemKeyValue[currentKeyExpr], currentKeyExpr);
            filterExpr.push(filterExprPart)
        }
        return filterExpr
    }
}
module.exports = Class.inherit({
    ctor: function(options) {
        options = options || {};
        this._selectionMode = options.mode;
        this._dataGetter = options.dataGetter;
        this._focusedItemIndex = -1;
        this._isAsyncMode = options.async;
        this._selectedItemKeys = options.selectedKeys || [];
        this._selectionFilter = void 0 === options.selectionFilter ? [] : options.selectionFilter;
        this._initSelectedItemKeyHash();
        this._selectedItems = [];
        this._selectionChangeHandler = options.onSelectionChanged || $.noop;
        this._clearItemKeys();
        this._maxFilterLengthInRequest = options.maxFilterLengthInRequest || 0;
        equalKeys = getEqualKeys(options.equalByReference)
    },
    _initSelectedItemKeyHash: function() {
        this._keyHashIndices = {}
    },
    _clearItemKeys: function() {
        this._addedItemKeys = [];
        this._removedItemKeys = []
    },
    validate: function() {
        var store = this._dataGetter.store();
        if (this._isAsyncMode && store && !store.key()) {
            throw errors.Error("E1042")
        }
    },
    getSelectedItemKeys: function() {
        if (this._isAsyncMode) {
            var d = $.Deferred(),
                dataSource = this._dataGetter.dataSource(),
                store = dataSource && dataSource.store(),
                key = store && store.key(),
                select = commonUtils.isString(key) ? [key] : key;
            this._loadFilteredData(this._selectionFilter, null, select).done(function(items) {
                var keys = items.map(function(item) {
                    return store.keyOf(item)
                });
                d.resolve(keys)
            }).fail(d.reject);
            return d.promise()
        } else {
            return this._selectedItemKeys
        }
    },
    getSelectedItems: function() {
        if (this._isAsyncMode) {
            return this._loadFilteredData(this._selectionFilter)
        } else {
            return this._selectedItems
        }
    },
    selectionFilter: function(value) {
        if (void 0 === value) {
            return this._selectionFilter
        }
        var filterIsChanged = this._selectionFilter !== value && JSON.stringify(this._selectionFilter) !== JSON.stringify(value);
        this._selectionFilter = value;
        filterIsChanged && this.onSelectionChanged()
    },
    selectedItemKeys: function(keys, preserve, isDeselect, isSelectAll) {
        var that = this;
        keys = keys || [];
        that.validate();
        if (that._isAsyncMode) {
            if (isSelectAll) {
                var filter = that._dataGetter.filter();
                if (!filter) {
                    that._selectionFilter = isDeselect ? [] : null
                } else {
                    that._addSelectionFilter(isDeselect, filter, false)
                }
            } else {
                if (!preserve) {
                    this._selectionFilter = []
                }
                for (var i = 0; i < keys.length; i++) {
                    this._addRemoveSelectedItemAsync(keys[i], isDeselect)
                }
            }
            that.onSelectionChanged();
            return $.Deferred().resolve()
        } else {
            var deferred = that._loadSelectedItems(keys, isDeselect, isSelectAll);
            deferred.done(function(items) {
                if (preserve) {
                    that._preserveSelectionUpdate(items, isDeselect)
                } else {
                    that._replaceSelectionUpdate(items)
                }
                that.onSelectionChanged()
            });
            return deferred
        }
    },
    _loadSelectedItems: function(itemKeys, isDeselect, isSelectAll) {
        var that = this,
            deferred = $.Deferred(),
            dataSource = this._dataGetter.dataSource(),
            store = dataSource && dataSource.store();
        if (!store) {
            deferred.resolve([]);
            return deferred
        }
        var keys = $.isArray(itemKeys) ? itemKeys : [itemKeys];
        if (!keys.length && !isSelectAll) {
            deferred.resolve([]);
            return deferred
        }
        if (isSelectAll && isDeselect && !that._dataGetter.filter()) {
            return that._clearSelection()
        }
        var selectionFilterCreator = new SelectionFilterCreator(dataSource.key(), keys, isSelectAll),
            combinedFilter = selectionFilterCreator.getCombinedFilter(that._dataGetter.filter());
        var deselectedItems = [];
        if (isDeselect) {
            deselectedItems = dataQuery(that._selectedItems).filter(combinedFilter).toArray()
        }
        var filteredItems = deselectedItems.length ? deselectedItems : dataSource.items();
        var filter = selectionFilterCreator.getFilter();
        if (filter) {
            filteredItems = dataQuery(filteredItems).filter(filter).toArray()
        }
        if (!isSelectAll && (deselectedItems.length || filteredItems.length === keys.length)) {
            deferred.resolve(filteredItems)
        } else {
            deferred = that._loadFilteredData(combinedFilter, selectionFilterCreator.getFunction())
        }
        return deferred
    },
    _clearSelection: function() {
        return this.selectedItemKeys([])
    },
    _loadFilteredData: function(filterExpr, filterFunction, select) {
        var dataSource = this._dataGetter.dataSource(),
            filterLength = encodeURI(JSON.stringify(filterExpr)).length,
            needLoadAllData = this._maxFilterLengthInRequest && filterLength > this._maxFilterLengthInRequest,
            deferred = $.Deferred(),
            loadOptions = {
                filter: needLoadAllData ? void 0 : filterExpr,
                select: needLoadAllData ? dataSource.select() : select || dataSource.select()
            };
        if (filterExpr && 0 === filterExpr.length) {
            deferred.resolve([])
        } else {
            this._dataGetter.load(loadOptions).done(function(items) {
                var filteredItems = items;
                if (needLoadAllData) {
                    filteredItems = dataQuery(filteredItems).filter(filterExpr).toArray()
                }
                if (filterFunction) {
                    filteredItems = dataQuery(filteredItems).filter(filterFunction).toArray()
                }
                deferred.resolve(filteredItems)
            }).fail($.proxy(deferred.reject, deferred))
        }
        return deferred
    },
    _preserveSelectionUpdate: function(items, isDeselect) {
        var dataSource = this._dataGetter.dataSource(),
            store = dataSource && dataSource.store();
        if (!store) {
            return
        }
        for (var i = 0; i < items.length; i++) {
            var item = items[i],
                key = store.keyOf(item);
            if (isDeselect) {
                this._removeSelectedItem(key)
            } else {
                this._addSelectedItem(item, key)
            }
        }
    },
    _indexOfSelectedItemKey: function(key) {
        var selectedIndex;
        if (commonUtils.isObject(key)) {
            selectedIndex = this._getSelectedIndexByKey(key)
        } else {
            selectedIndex = this._getSelectedIndexByHash(key)
        }
        return selectedIndex
    },
    _getSelectedIndexByKey: function(key) {
        var selectedItemKeys = this._selectedItemKeys;
        for (var index = 0; index < selectedItemKeys.length; index++) {
            if (equalKeys(selectedItemKeys[index], key)) {
                return index
            }
        }
        return -1
    },
    _getSelectedIndexByHash: function(key) {
        var indices = this._keyHashIndices[key];
        return indices && indices[0] >= 0 ? indices[0] : -1
    },
    _shiftSelectedKeyIndices: function(keyIndex) {
        for (var currentKeyIndex = keyIndex; currentKeyIndex < this._selectedItemKeys.length; currentKeyIndex++) {
            var currentKey = this._selectedItemKeys[currentKeyIndex],
                currentKeyHash = getKeyHash(currentKey),
                currentKeyIndices = this._keyHashIndices[currentKeyHash];
            if (!currentKeyIndices) {
                continue
            }
            for (var i = 0; i < currentKeyIndices.length; i++) {
                if (currentKeyIndices[i] > keyIndex) {
                    currentKeyIndices[i]--
                }
            }
        }
    },
    _addFilterOperator: function(selectionFilter, filterOperator) {
        if (selectionFilter.length > 1 && commonUtils.isString(selectionFilter[1]) && selectionFilter[1] !== filterOperator) {
            selectionFilter = [selectionFilter]
        }
        if (selectionFilter.length) {
            selectionFilter.push(filterOperator)
        }
        return selectionFilter
    },
    _hasSameFilter: function(selectionFilter, currentFilter) {
        return this._findSubFilter(selectionFilter, currentFilter) >= 0
    },
    _removeInvertedFilter: function(selectionFilter, isDeselect, filter) {
        filter = isDeselect ? filter : ["!", filter];
        var filterIndex = this._findSubFilter(selectionFilter, filter);
        if (JSON.stringify(filter) === JSON.stringify(selectionFilter)) {
            selectionFilter.splice(0, selectionFilter.length);
            return true
        }
        if (filterIndex >= 0) {
            if (filterIndex > 0) {
                selectionFilter.splice(filterIndex - 1, 2)
            } else {
                selectionFilter.splice(filterIndex, 2)
            }
            return true
        }
        return false
    },
    _denormalizeFilter: function(filter) {
        if (filter && commonUtils.isString(filter[0])) {
            filter = [filter]
        }
        return filter
    },
    _normalizeFilter: function(filter) {
        if (filter && 1 === filter.length) {
            filter = filter[0]
        }
        return filter
    },
    _addSelectionFilter: function(isDeselect, filter, isUnique) {
        var that = this,
            needAddFilter = true,
            currentFilter = isDeselect ? ["!", filter] : filter,
            selectionFilter = that._selectionFilter || [];
        selectionFilter = that._denormalizeFilter(selectionFilter);
        if (selectionFilter && selectionFilter.length) {
            if (that._hasSameFilter(selectionFilter, currentFilter)) {
                return
            }
            if (that._removeInvertedFilter(selectionFilter, isDeselect, filter)) {
                needAddFilter = !isUnique
            }
            if (needAddFilter) {
                selectionFilter = that._addFilterOperator(selectionFilter, isDeselect ? "and" : "or")
            }
        }
        if (needAddFilter) {
            selectionFilter.push(currentFilter)
        }
        selectionFilter = that._normalizeFilter(selectionFilter);
        that._selectionFilter = selectionFilter
    },
    _findSubFilter: function(selectionFilter, filter) {
        if (!selectionFilter) {
            return -1
        }
        var filterString = JSON.stringify(filter);
        for (var index = 0; index < selectionFilter.length; index++) {
            var subFilter = selectionFilter[index];
            if (subFilter && JSON.stringify(subFilter) === filterString) {
                return index
            }
        }
        return -1
    },
    _addRemoveSelectedItemAsync: function(key, isDeselect) {
        var dataSource = this._dataGetter.dataSource(),
            keyField = dataSource && dataSource.store().key(),
            filter = [keyField, "=", key];
        if ($.isArray(keyField)) {
            filter = [];
            for (var i = 0; i < keyField.length; i++) {
                filter.push([keyField[i], "=", key[keyField[i]]]);
                if (i !== keyField.length - 1) {
                    filter.push("and")
                }
            }
        }
        this._addSelectionFilter(isDeselect, filter, true)
    },
    _addSelectedItem: function(itemData, key) {
        if (this._isAsyncMode) {
            this._addRemoveSelectedItemAsync(key)
        } else {
            var keyHash = getKeyHash(key);
            if (this._indexOfSelectedItemKey(keyHash) === -1) {
                if (!commonUtils.isObject(keyHash)) {
                    this._keyHashIndices[keyHash] = [this._selectedItemKeys.length]
                }
                this._selectedItemKeys.push(key);
                this._addedItemKeys.push(key);
                this._selectedItems.push(itemData)
            }
        }
    },
    _removeSelectedItem: function(key) {
        if (this._isAsyncMode) {
            this._addRemoveSelectedItemAsync(key, true)
        } else {
            var keyHash = getKeyHash(key),
                keyIndex = this._indexOfSelectedItemKey(keyHash);
            if (keyIndex < 0) {
                return
            }
            this._selectedItemKeys.splice(keyIndex, 1);
            this._removedItemKeys.push(key);
            this._selectedItems.splice(keyIndex, 1);
            if (commonUtils.isObject(keyHash)) {
                return
            }
            var keyIndices = this._keyHashIndices[keyHash];
            if (!keyIndices) {
                return
            }
            keyIndices.shift();
            if (!keyIndices.length) {
                delete this._keyHashIndices[keyHash]
            }
            this._shiftSelectedKeyIndices(keyIndex)
        }
    },
    _replaceSelectionUpdate: function(items) {
        var dataSource = this._dataGetter.dataSource(),
            store = dataSource && dataSource.store(),
            internalKeys = [];
        if (!store) {
            return
        }
        for (var i = 0; i < items.length; i++) {
            var item = items[i],
                key = store.keyOf(item);
            internalKeys.push(key)
        }
        this._setSelectedItems(internalKeys, items)
    },
    _setSelectedItems: function(keys, items) {
        if (this._isAsyncMode) {
            this._selectionFilter = null;
            for (var i = 0; i < keys.length; i++) {
                this._addRemoveSelectedItemAsync(keys[i])
            }
        } else {
            this._updateRemovedItemKeys(keys);
            this._updateAddedItemKeys(keys);
            this._initSelectedItemKeyHash();
            this._updateSelectedItemKeyHash(keys);
            this._selectedItemKeys = keys;
            this._selectedItems = items
        }
    },
    _updateRemovedItemKeys: function(keys) {
        var oldSelectedKeys = this._selectedItemKeys;
        for (var i = 0; i < oldSelectedKeys.length; i++) {
            var hash = getKeyHash(oldSelectedKeys[i]);
            if (this._needRemoveItemKey(keys, hash)) {
                this._removedItemKeys.push(oldSelectedKeys[i])
            }
        }
    },
    _needRemoveItemKey: function(keys, hash) {
        var hashIndices = this._keyHashIndices;
        for (var i = 0; i < keys.length; i++) {
            var keyHash = getKeyHash(keys[i]);
            if (hashIndices[keyHash] === hashIndices[hash]) {
                return false
            }
        }
        return true
    },
    _updateAddedItemKeys: function(keys) {
        for (var i = 0; i < keys.length; i++) {
            var keyHash = getKeyHash(keys[i]);
            if (void 0 === this._keyHashIndices[keyHash]) {
                this._addedItemKeys.push(keys[i])
            }
        }
    },
    _updateSelectedItemKeyHash: function(keys) {
        for (var i = 0; i < keys.length; i++) {
            var keyHash = getKeyHash(keys[i]);
            if (!commonUtils.isObject(keyHash)) {
                this._keyHashIndices[keyHash] = this._keyHashIndices[keyHash] || [];
                var keyIndices = this._keyHashIndices[keyHash];
                keyIndices.push(i)
            }
        }
    },
    onSelectionChanged: function() {
        this._selectionChangeHandler(this._selectedItems, this._selectedItemKeys, this._addedItemKeys, this._removedItemKeys);
        this._clearItemKeys()
    },
    changeItemSelection: function(itemIndex, keys) {
        var isSelectedItemsChanged, items = this._dataGetter.items(),
            item = items[itemIndex];
        if (!this.isSelectable() || !this.isDataItem(item)) {
            return false
        }
        var itemData = this._dataGetter.getItemData(item),
            itemKey = this._dataGetter.store().keyOf(itemData);
        keys = keys || {};
        if (keys.shift && "multiple" === this._selectionMode && this._focusedItemIndex >= 0) {
            isSelectedItemsChanged = this.changeItemSelectionWhenShiftKeyPressed(itemIndex, items)
        } else {
            if (keys.control) {
                this._resetItemSelectionWhenShiftKeyPressed();
                var isSelected = this.isItemSelected(this._isAsyncMode ? itemData : itemKey);
                if ("single" === this._selectionMode) {
                    this.clearSelectedItems()
                }
                if (isSelected) {
                    this._removeSelectedItem(itemKey)
                } else {
                    this._addSelectedItem(itemData, itemKey)
                }
                isSelectedItemsChanged = true
            } else {
                this._resetItemSelectionWhenShiftKeyPressed();
                if (1 !== this._selectedItemKeys.length || !equalKeys(this._selectedItemKeys[0], itemKey)) {
                    this._setSelectedItems([itemKey], [itemData]);
                    isSelectedItemsChanged = true
                }
            }
        }
        if (isSelectedItemsChanged) {
            this._focusedItemIndex = itemIndex;
            this.onSelectionChanged();
            return true
        }
    },
    isDataItem: function(item) {
        return this._dataGetter.isDataItem(item)
    },
    isSelectable: function() {
        return "single" === this._selectionMode || "multiple" === this._selectionMode
    },
    isItemSelected: function(arg) {
        if (this._isAsyncMode) {
            var selectionFilter = this._selectionFilter;
            if (!selectionFilter) {
                return true
            }
            return !!dataQuery([arg]).filter(selectionFilter).toArray().length
        } else {
            var keyHash = getKeyHash(arg);
            var index = this._indexOfSelectedItemKey(keyHash);
            return index !== -1
        }
    },
    _resetItemSelectionWhenShiftKeyPressed: function() {
        delete this._shiftFocusedItemIndex
    },
    changeItemSelectionWhenShiftKeyPressed: function(itemIndex, items) {
        var itemIndexStep, index, isSelectedItemsChanged = false,
            store = this._dataGetter.store(),
            key = store.keyOf(items[this._focusedItemIndex].data),
            isFocusedItemSelected = items[this._focusedItemIndex] && this.isItemSelected(key);
        if (!commonUtils.isDefined(this._shiftFocusedItemIndex)) {
            this._shiftFocusedItemIndex = this._focusedItemIndex
        }
        var data, itemKey;
        if (this._shiftFocusedItemIndex !== this._focusedItemIndex) {
            itemIndexStep = this._focusedItemIndex < this._shiftFocusedItemIndex ? 1 : -1;
            for (index = this._focusedItemIndex; index !== this._shiftFocusedItemIndex; index += itemIndexStep) {
                if (this.isDataItem(items[index])) {
                    itemKey = store.keyOf(this._dataGetter.getItemData(items[index]));
                    this._removeSelectedItem(itemKey);
                    isSelectedItemsChanged = true
                }
            }
        }
        if (itemIndex !== this._shiftFocusedItemIndex) {
            itemIndexStep = itemIndex < this._shiftFocusedItemIndex ? 1 : -1;
            for (index = itemIndex; index !== this._shiftFocusedItemIndex; index += itemIndexStep) {
                if (this.isDataItem(items[index])) {
                    data = this._dataGetter.getItemData(items[index]);
                    itemKey = store.keyOf(data);
                    this._addSelectedItem(data, itemKey);
                    isSelectedItemsChanged = true
                }
            }
        }
        if (this.isDataItem(items[this._focusedItemIndex]) && !isFocusedItemSelected) {
            data = this._dataGetter.getItemData(items[index]);
            itemKey = store.keyOf(data);
            this._addSelectedItem(data, itemKey);
            isSelectedItemsChanged = true
        }
        return isSelectedItemsChanged
    },
    clearSelectedItems: function() {
        this._setSelectedItems([], [])
    },
    selectAll: function(isOnePage) {
        if (isOnePage) {
            return this._onePageSelectAll(false)
        } else {
            return this.selectedItemKeys([], true, false, true)
        }
    },
    deselectAll: function(isOnePage) {
        if (isOnePage) {
            return this._onePageSelectAll(true)
        } else {
            return this.selectedItemKeys([], true, true, true)
        }
    },
    _onePageSelectAll: function(isDeselect) {
        var items = this._dataGetter.items();
        for (var i = 0; i < items.length; i++) {
            var item = items[i],
                itemData = this._dataGetter.getItemData(item),
                itemKey = this._dataGetter.store().keyOf(itemData),
                isSelected = this.isItemSelected(itemKey);
            if (!isSelected && !isDeselect) {
                this._addSelectedItem(itemData, itemKey)
            }
            if (isSelected && isDeselect) {
                this._removeSelectedItem(itemKey)
            }
        }
        this.onSelectionChanged();
        return $.Deferred().resolve()
    },
    _isLastSubFilter: function(selectionFilter, filter) {
        if (selectionFilter && filter) {
            return this._findSubFilter(selectionFilter, filter) === selectionFilter.length - 1
        }
        return false
    },
    _isSelectAllAsync: function() {
        var filter = this._dataGetter.filter(),
            selectionFilter = this._selectionFilter;
        if (filter && filter.length && selectionFilter && selectionFilter.length) {
            selectionFilter = this._denormalizeFilter(selectionFilter);
            if (this._isLastSubFilter(selectionFilter, filter)) {
                return true
            }
            if (this._isLastSubFilter(selectionFilter, ["!", filter])) {
                return false
            }
        }
        if (!selectionFilter) {
            return true
        }
        if (selectionFilter && 0 === selectionFilter.length) {
            return false
        }
        return
    },
    _isSelectAllOnePage: function() {
        var items = this._dataGetter.items(),
            dataSource = this._dataGetter.dataSource(),
            store = dataSource && dataSource.store(),
            hasSelectedItems = false,
            hasUnselectedItems = false;
        for (var i = 0; i < items.length; i++) {
            var item = items[i];
            var itemData = this._dataGetter.getItemData(item);
            var key = store.keyOf(itemData);
            if (this._dataGetter.isDataItem(item)) {
                if (this.isItemSelected(key)) {
                    hasSelectedItems = true
                } else {
                    hasUnselectedItems = true
                }
            }
        }
        if (hasSelectedItems) {
            return !hasUnselectedItems ? true : void 0
        } else {
            return false
        }
    },
    isSelectAll: function(isOnePage) {
        if (this._isAsyncMode) {
            return this._isSelectAllAsync()
        } else {
            if (isOnePage) {
                return this._isSelectAllOnePage()
            } else {
                var items = this._dataGetter.items(),
                    selectedItems = this._selectedItems;
                var dataFilter = this._dataGetter.filter();
                if (dataFilter) {
                    selectedItems = dataQuery(selectedItems).filter(dataFilter).toArray()
                }
                var selectedItemsLength = selectedItems.length;
                if (!selectedItemsLength) {
                    return this._isSelectedItem(items)
                }
                if (selectedItemsLength >= this._dataGetter.totalCount()) {
                    return true
                }
            }
        }
    },
    _isSelectedItem: function(items) {
        for (var i = 0; i < items.length; i++) {
            if (items[i].selected) {
                return
            }
        }
        return false
    }
});
