/**
 * DevExtreme (ui/selection/selection.base.js)
 * Version: 16.2.1-16314
 * Build date: Wed Nov 09 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 */
"use strict";
var Class = require("../../core/class"),
    $ = require("jquery"),
    commonUtils = require("../../core/utils/common");
module.exports = Class.inherit({
    ctor: function(options) {
        options = options || {};
        this._options = $.extend(this._getDefaultOptions(), options);
        this._initItems();
        this._initActions()
    },
    _getDefaultOptions: function() {
        return {
            mode: "multiple",
            keyExpr: "id",
            selectedExpr: "selected",
            selectableExpr: "selectable",
            items: [],
            onSelectionChanged: null
        }
    },
    option: function(name, value) {
        if (!name) {
            return this._options
        }
        if (!value) {
            return this._options[name]
        }
        var oldValue = this._options[name];
        this._options[name] = value;
        this._optionChanged(name, oldValue, value)
    },
    _getKeyHash: function(key) {
        if (commonUtils.isObject(key) || commonUtils.isArray(key)) {
            try {
                var keyHash = JSON.stringify(key);
                return "{}" === keyHash ? key : keyHash
            } catch (e) {
                return key
            }
        }
        return key
    },
    _clearKeyMap: function() {
        this._itemKeyMap = []
    },
    _clearSelection: function() {
        this._selectedItemKeys = []
    },
    _removeSelectedItemKey: function(key) {
        var keyHash = this._getKeyHash(key),
            keyIndex = this._selectedItemKeys.indexOf(keyHash);
        if (keyIndex === -1) {
            return
        }
        this._selectedItemKeys.splice(keyIndex, 1)
    },
    _addSelectedItemKey: function(key) {
        var keyHash = this._getKeyHash(key),
            keyIndex = this._selectedItemKeys.indexOf(keyHash);
        if (keyIndex !== -1) {
            return
        }
        this._selectedItemKeys.push(keyHash)
    },
    _initActions: function() {
        this._onSelectionChanged = this.option("onSelectionChanged") || $.noop
    },
    _initItems: function() {
        var items = this.option("items"),
            that = this;
        this._clearKeyMap();
        this._clearSelection();
        $.each(items, function(_, item) {
            var keyValue = that._getValueByExpr(item, "key"),
                hash = that._getKeyHash(keyValue);
            that._itemKeyMap[hash] = item;
            if (that._isItemSelected(item)) {
                that._addSelectedItemKey(keyValue)
            }
        })
    },
    _getFieldNameByExpr: function(expr, item) {
        var exprOptionValue = this.option(expr + "Expr"),
            fieldName = commonUtils.isString(exprOptionValue) ? exprOptionValue : exprOptionValue(item);
        return fieldName
    },
    _isItem: function(object) {
        return this.option("items").indexOf(object) !== -1
    },
    _getItemByKey: function(key) {
        if (this._isItem(key)) {
            return key
        }
        var hash = this._getKeyHash(key);
        return this._itemKeyMap[hash] || {}
    },
    _getValueByExpr: function(item, expr) {
        var fieldName = this._getFieldNameByExpr(expr, item);
        return item[fieldName]
    },
    _setValueByExpr: function(item, expr, value) {
        var fieldName = this._getFieldNameByExpr(expr, item);
        item[fieldName] = value
    },
    _isItemSelectionEnabled: function(item) {
        return this._isItemSelectable(item) && "none" !== this.option("mode")
    },
    _toggleSelection: function(keys, state) {
        keys = Array.isArray(keys) ? keys : [keys];
        $.each(keys, $.proxy(function(_, key) {
            var item = this._getItemByKey(key),
                isItemSelected = this._isItemSelected(item);
            if (!this._isItemSelectionEnabled(item) || isItemSelected === state) {
                return
            }
            this._setValueByExpr(item, "selected", state);
            if (this._isItem(key)) {
                key = this._getValueByExpr(item, "key")
            }
            state ? this._addSelectedItemKey(key) : this._removeSelectedItemKey(key);
            this._onSelectionChanged({
                itemData: item,
                oldValue: isItemSelected,
                value: state
            })
        }, this))
    },
    _isItemSelectable: function(item) {
        var value = this._getValueByExpr(item, "selectable");
        if (commonUtils.isDefined(value)) {
            return value
        }
        return true
    },
    _isItemSelected: function(item) {
        return !!this._getValueByExpr(item, "selected")
    },
    _optionChanged: function(name, old, value) {
        switch (name) {
            case "items":
                this._initItems();
                break;
            case "onSelectionChanged":
                this._initActions()
        }
    },
    getSelectedItemKeys: function() {
        return this._selectedItemKeys
    },
    select: function(keys) {
        if ("single" === this.option("mode")) {
            this.deselectAll()
        }
        this._toggleSelection(keys, true)
    },
    deselect: function(keys) {
        this._toggleSelection(keys, false)
    },
    selectAll: function() {
        $.each(this.option("items"), $.proxy(function(_, item) {
            this._toggleSelection(item, true)
        }, this))
    },
    deselectAll: function() {
        this._toggleSelection(this.getSelectedItemKeys().slice(), false)
    },
    isAllSelected: function() {
        return this.option("items").length === this.getSelectedItemKeys().length
    }
});
